"""
This module defines interfaces for requests and responses used in Connexion for authentication,
validation, serialization, etc.
"""
import typing as t
from collections import defaultdict

from multipart.multipart import parse_options_header
from starlette.datastructures import UploadFile
from starlette.requests import Request as StarletteRequest
from werkzeug import Request as WerkzeugRequest

from connexion.http_facts import FORM_CONTENT_TYPES
from connexion.utils import is_json_mimetype


class _RequestInterface:
    @property
    def context(self) -> t.Dict[str, t.Any]:
        """The connexion context of the current request cycle."""
        raise NotImplementedError

    @property
    def content_type(self) -> str:
        raise NotImplementedError

    @property
    def mimetype(self) -> str:
        raise NotImplementedError

    @property
    def path_params(self) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @property
    def query_params(self) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    def form(self) -> t.Union[t.Dict[str, t.Any], t.Awaitable[t.Dict[str, t.Any]]]:
        raise NotImplementedError

    def files(self) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    def get_body(self) -> t.Any:
        raise NotImplementedError


class WSGIRequest(_RequestInterface):
    def __init__(
        self, werkzeug_request: WerkzeugRequest, uri_parser=None, view_args=None
    ):
        self._werkzeug_request = werkzeug_request
        self.uri_parser = uri_parser
        self.view_args = view_args

        self._context = None
        self._path_params = None
        self._query_params = None
        self._form = None
        self._body = None

    @property
    def context(self):
        if self._context is None:
            scope = self.environ["asgi.scope"]
            extensions = scope.setdefault("extensions", {})
            self._context = extensions.setdefault("connexion_context", {})
        return self._context

    @property
    def content_type(self) -> str:
        return self._werkzeug_request.content_type or "application/octet-stream"

    @property
    def mimetype(self) -> str:
        return self._werkzeug_request.mimetype

    @property
    def path_params(self):
        if self._path_params is None:
            self._path_params = self.uri_parser.resolve_path(self.view_args)
        return self._path_params

    @property
    def query_params(self):
        if self._query_params is None:
            query_params = {k: self.args.getlist(k) for k in self.args}
            self._query_params = self.uri_parser.resolve_query(query_params)
        return self._query_params

    def form(self):
        if self._form is None:
            form = self._werkzeug_request.form.to_dict(flat=False)
            self._form = self.uri_parser.resolve_form(form)
        return self._form

    def files(self):
        return self._werkzeug_request.files.to_dict(flat=False)

    def get_body(self):
        """Get body based on content type"""
        if self._body is None:
            if is_json_mimetype(self.content_type):
                self._body = self.get_json(silent=True)
            elif self.mimetype in FORM_CONTENT_TYPES:
                self._body = self.form()
            else:
                # Return explicit None instead of empty bytestring so it is handled as null downstream
                self._body = self.get_data() or None
        return self._body

    def __getattr__(self, item):
        return getattr(self._werkzeug_request, item)


class ASGIRequest(_RequestInterface):
    """Wraps starlette Request so it can easily be extended."""

    def __init__(self, *args, uri_parser=None, **kwargs):
        self._starlette_request = StarletteRequest(*args, **kwargs)
        self.uri_parser = uri_parser

        self._context = None
        self._mimetype = None
        self._path_params = None
        self._query_params = None
        self._form = None
        self._files = None

    @property
    def context(self):
        if self._context is None:
            extensions = self.scope.setdefault("extensions", {})
            self._context = extensions.setdefault("connexion_context", {})
        return self._context

    @property
    def content_type(self):
        return self.headers.get("content-type", "application/octet-stream")

    @property
    def mimetype(self):
        if not self._mimetype:
            mimetype, _ = parse_options_header(self.content_type)
            self._mimetype = mimetype.decode()
        return self._mimetype

    @property
    def path_params(self) -> t.Dict[str, t.Any]:
        if self._path_params is None:
            self._path_params = self.uri_parser.resolve_path(
                self._starlette_request.path_params
            )
        return self._path_params

    @property
    def query_params(self):
        if self._query_params is None:
            args = self._starlette_request.query_params
            query_params = {k: args.getlist(k) for k in args}
            self._query_params = self.uri_parser.resolve_query(query_params)
        return self._query_params

    async def form(self):
        if self._form is None:
            await self._split_form_files()
        return self._form

    async def files(self):
        if self._files is None:
            await self._split_form_files()
        return self._files

    async def _split_form_files(self):
        form_data = await self._starlette_request.form()

        files = defaultdict(list)
        form = defaultdict(list)
        for k, v in form_data.multi_items():
            if isinstance(v, UploadFile):
                files[k].append(v)
            else:
                form[k].append(v)

        self._files = files
        self._form = self.uri_parser.resolve_form(form)

    async def json(self):
        try:
            return await self._starlette_request.json()
        except ValueError:
            return None

    async def get_body(self):
        if is_json_mimetype(self.content_type):
            return await self.json()
        elif self.mimetype in FORM_CONTENT_TYPES:
            return await self.form()
        else:
            # Return explicit None instead of empty bytestring so it is handled as null downstream
            return await self.body() or None

    def __getattr__(self, item):
        return getattr(self._starlette_request, item)


class ConnexionResponse:
    """Connexion interface for a response."""

    def __init__(
        self,
        status_code=200,
        mimetype=None,
        content_type=None,
        body=None,
        headers=None,
        is_streamed=False,
    ):
        self.status_code = status_code
        self.mimetype = mimetype
        self.content_type = content_type
        self.body = body
        self.headers = headers or {}
        self.headers.update({"Content-Type": content_type})
        self.is_streamed = is_streamed



import logging
import typing as t

from jsonschema import ValidationError, draft4_format_checker
from starlette.datastructures import Headers, UploadFile
from starlette.formparsers import FormParser, MultiPartParser
from starlette.types import Scope

from connexion.exceptions import BadRequestProblem, ExtraParameterProblem
from connexion.json_schema import Draft4RequestValidator, format_error_with_path
from connexion.uri_parsing import AbstractURIParser
from connexion.validators import AbstractRequestBodyValidator

logger = logging.getLogger("connexion.validators.form_data")


class FormDataValidator(AbstractRequestBodyValidator):
    """Request body validator for form content types."""

    def __init__(
        self,
        *,
        schema: dict,
        required=False,
        nullable=False,
        encoding: str,
        strict_validation: bool,
        uri_parser: t.Optional[AbstractURIParser] = None,
    ) -> None:
        super().__init__(
            schema=schema,
            required=required,
            nullable=nullable,
            encoding=encoding,
            strict_validation=strict_validation,
        )
        self._uri_parser = uri_parser

    @property
    def _validator(self):
        return Draft4RequestValidator(
            self._schema, format_checker=draft4_format_checker
        )

    @property
    def _form_parser_cls(self):
        return FormParser

    async def _parse(self, stream: t.AsyncGenerator[bytes, None], scope: Scope) -> dict:
        headers = Headers(scope=scope)
        form_parser = self._form_parser_cls(headers, stream)
        data = await form_parser.parse()

        if self._uri_parser is not None:
            # Don't parse file_data
            form_data = {}
            file_data = {}
            for k, v in data.items():
                if isinstance(v, str):
                    form_data[k] = data.getlist(k)
                elif isinstance(v, UploadFile):
                    # Replace files with empty strings for validation
                    file_data[k] = ""

            data = self._uri_parser.resolve_form(form_data)
            # Add the files again
            data.update(file_data)
        else:
            data = {k: data.getlist(k) for k in data}

        return data

    def _validate(self, data: dict) -> None:
        if self._strict_validation:
            self._validate_params_strictly(data)

        try:
            self._validator.validate(data)
        except ValidationError as exception:
            error_path_msg = format_error_with_path(exception=exception)
            logger.error(
                f"Validation error: {exception.message}{error_path_msg}",
                extra={"validator": "body"},
            )
            raise BadRequestProblem(detail=f"{exception.message}{error_path_msg}")

    def _validate_params_strictly(self, data: dict) -> None:
        form_params = data.keys()
        spec_params = self._schema.get("properties", {}).keys()
        errors = set(form_params).difference(set(spec_params))
        if errors:
            raise ExtraParameterProblem(param_type="formData", extra_params=errors)


class MultiPartFormDataValidator(FormDataValidator):
    @property
    def _form_parser_cls(self):
        return MultiPartParser



import dataclasses
import logging
import pathlib
import typing as t
from dataclasses import dataclass, field

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion import utils
from connexion.handlers import ResolverErrorHandler
from connexion.jsonifier import Jsonifier
from connexion.middleware.abstract import SpecMiddleware
from connexion.middleware.context import ContextMiddleware
from connexion.middleware.exceptions import ExceptionMiddleware
from connexion.middleware.lifespan import Lifespan, LifespanMiddleware
from connexion.middleware.request_validation import RequestValidationMiddleware
from connexion.middleware.response_validation import ResponseValidationMiddleware
from connexion.middleware.routing import RoutingMiddleware
from connexion.middleware.security import SecurityMiddleware
from connexion.middleware.swagger_ui import SwaggerUIMiddleware
from connexion.resolver import Resolver
from connexion.uri_parsing import AbstractURIParser

logger = logging.getLogger(__name__)


@dataclass
class _Options:
    """
    Connexion provides a lot of parameters for the user to configure the app / middleware of
    application.

    This class provides a central place to parse these parameters a mechanism to update them.
    Application level arguments can be provided when instantiating the application / middleware,
    after which they can be overwritten on an API level.

    The defaults should only be set in this class, and set to None in the signature of user facing
    methods. This is necessary for this class to be able to differentiate between missing and
    falsy arguments.
    """

    arguments: t.Optional[dict] = None
    auth_all_paths: t.Optional[bool] = False
    jsonifier: t.Optional[Jsonifier] = None
    pythonic_params: t.Optional[bool] = False
    resolver: t.Optional[t.Union[Resolver, t.Callable]] = None
    resolver_error: t.Optional[int] = None
    resolver_error_handler: t.Optional[t.Callable] = field(init=False)
    strict_validation: t.Optional[bool] = False
    swagger_ui_options: t.Optional[dict] = None
    uri_parser_class: t.Optional[AbstractURIParser] = None
    validate_responses: t.Optional[bool] = False
    validator_map: t.Optional[dict] = None

    def __post_init__(self):
        self.resolver = (
            Resolver(self.resolver) if callable(self.resolver) else self.resolver
        )
        self.resolver_error_handler = self._resolver_error_handler_factory()

    def _resolver_error_handler_factory(
        self,
    ) -> t.Optional[t.Callable[[], ResolverErrorHandler]]:
        """Returns a factory to create a ResolverErrorHandler."""
        if self.resolver_error is not None:

            def resolver_error_handler(*args, **kwargs) -> ResolverErrorHandler:
                return ResolverErrorHandler(self.resolver_error, *args, **kwargs)

            return resolver_error_handler
        return None

    def replace(self, **changes) -> "_Options":
        """Update mechanism to overwrite the options. None values are discarded.

        :param changes: Arguments accepted by the __init__ method of this class.

        :return: An new _Options object with updated arguments.
        """
        changes = {key: value for key, value in changes.items() if value is not None}
        return dataclasses.replace(self, **changes)


class ConnexionMiddleware:
    """The main Connexion middleware, which wraps a list of specialized middlewares around the
    provided application."""

    default_middlewares = [
        ExceptionMiddleware,
        SwaggerUIMiddleware,
        RoutingMiddleware,
        SecurityMiddleware,
        RequestValidationMiddleware,
        ResponseValidationMiddleware,
        ContextMiddleware,
        LifespanMiddleware,
    ]

    def __init__(
        self,
        app: ASGIApp,
        *,
        import_name: t.Optional[str] = None,
        lifespan: t.Optional[Lifespan] = None,
        middlewares: t.Optional[list] = None,
        specification_dir: t.Union[pathlib.Path, str] = "",
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[dict] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
    ):
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionmMiddleware.default_middlewares`
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A :class:`options.ConnexionOptions` instance with configuration
            options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        """
        import_name = import_name or str(pathlib.Path.cwd())
        self.root_path = utils.get_root_path(import_name)

        self.specification_dir = self.ensure_absolute(specification_dir)

        if middlewares is None:
            middlewares = self.default_middlewares
        self.app, self.apps = self._apply_middlewares(
            app, middlewares, lifespan=lifespan
        )

        self.options = _Options(
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            swagger_ui_options=swagger_ui_options,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
        )

        self.extra_files: t.List[str] = []

    def ensure_absolute(self, path: t.Union[str, pathlib.Path]) -> pathlib.Path:
        """Ensure that a path is absolute. If the path is not absolute, it is assumed to relative
        to the application root path and made absolute."""
        path = pathlib.Path(path)
        if path.is_absolute():
            return path
        else:
            return self.root_path / path

    def _apply_middlewares(
        self, app: ASGIApp, middlewares: t.List[t.Type[ASGIApp]], **kwargs
    ) -> t.Tuple[ASGIApp, t.Iterable[ASGIApp]]:
        """Apply all middlewares to the provided app.

        :param app: App to wrap in middlewares.
        :param middlewares: List of middlewares to wrap around app. The list should be ordered
                            from outer to inner middleware.

        :return: Tuple of the outer middleware wrapping the application and a list of the wrapped
            middlewares, including the wrapped application.
        """
        # Include the wrapped application in the returned list.
        apps = [app]
        for middleware in reversed(middlewares):
            app = middleware(app, **kwargs)  # type: ignore
            apps.append(app)
        return app, list(reversed(apps))

    def add_api(
        self,
        specification: t.Union[pathlib.Path, str, dict],
        *,
        base_path: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[dict] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        **kwargs,
    ) -> t.Any:
        """
        Register een API represented by a single OpenAPI specification on this middleware.
        Multiple APIs can be registered on a single middleware.

        :param specification: OpenAPI specification. Can be provided either as dict, or as path
            to file.
        :param base_path: Base path to host the API. This overrides the basePath / servers in the
            specification.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A :class:`options.ConnexionOptions` instance with configuration
            options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`
        :param kwargs: Additional keyword arguments to pass to the `add_api` method of the managed
            middlewares. This can be used to pass arguments to middlewares added beyond the default
            ones.

        :return: The Api registered on the wrapped application.
        """
        if isinstance(specification, dict):
            specification = specification
        else:
            specification = t.cast(pathlib.Path, self.specification_dir / specification)
            # Add specification as file to watch for reloading
            if pathlib.Path.cwd() in specification.parents:
                self.extra_files.append(
                    str(specification.relative_to(pathlib.Path.cwd()))
                )

        options = self.options.replace(
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            swagger_ui_options=swagger_ui_options,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
        )

        for app in self.apps:
            if isinstance(app, SpecMiddleware):
                api = app.add_api(
                    specification,
                    base_path=base_path,
                    **options.__dict__,
                    **kwargs,
                )

        # Api registered on the inner application.
        return api

    def add_error_handler(
        self, code_or_exception: t.Union[int, t.Type[Exception]], function: t.Callable
    ) -> None:
        for app in self.apps:
            if isinstance(app, ExceptionMiddleware):
                app.add_exception_handler(code_or_exception, function)

    def run(self, import_string: str = None, **kwargs):
        """Run the application using uvicorn.

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param kwargs: kwargs to pass to `uvicorn.run`.
        """
        try:
            import uvicorn
        except ImportError:
            raise RuntimeError(
                "uvicorn is not installed. Please install connexion using the uvicorn extra "
                "(connexion[uvicorn])"
            )

        logger.warning(
            f"`{self.__class__.__name__}.run` is optimized for development. "
            "For production, run using a dedicated ASGI server."
        )

        app: t.Union[str, ConnexionMiddleware]
        if import_string is not None:
            app = import_string
            kwargs.setdefault("reload", True)
            kwargs["reload_includes"] = self.extra_files + kwargs.get(
                "reload_includes", []
            )
        else:
            app = self

        uvicorn.run(app, **kwargs)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.app(scope, receive, send)



"""
This module defines an AbstractApp, which defines a standardized user interface for a Connexion
application.
"""
import abc
import pathlib
import typing as t

from starlette.testclient import TestClient
from starlette.types import Receive, Scope, Send

from connexion.jsonifier import Jsonifier
from connexion.middleware import ConnexionMiddleware, SpecMiddleware
from connexion.middleware.lifespan import Lifespan
from connexion.resolver import Resolver
from connexion.uri_parsing import AbstractURIParser


class AbstractApp:
    """
    Abstract class for a Connexion Application. A Connexion Application provides an interface for a
    framework application wrapped by Connexion Middleware. Since its main function is to provide an
    interface, it delegates most of the work to the middleware and framework application.
    """

    middleware_app: SpecMiddleware
    """
    The application wrapped by the ConnexionMiddleware, which in its turn wraps the framework
    application.
    """

    def __init__(
        self,
        import_name: str,
        *,
        lifespan: t.Optional[Lifespan] = None,
        middlewares: t.Optional[list] = None,
        specification_dir: t.Union[pathlib.Path, str] = "",
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[dict] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
    ) -> None:
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionmMiddleware.default_middlewares`
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A :class:`options.ConnexionOptions` instance with configuration
            options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        """
        self.middleware = ConnexionMiddleware(
            self.middleware_app,
            import_name=import_name,
            lifespan=lifespan,
            middlewares=middlewares,
            specification_dir=specification_dir,
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            swagger_ui_options=swagger_ui_options,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
        )

    def add_api(
        self,
        specification: t.Union[pathlib.Path, str, dict],
        *,
        base_path: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[dict] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        **kwargs,
    ) -> t.Any:
        """
        Register een API represented by a single OpenAPI specification on this application.
        Multiple APIs can be registered on a single application.

        :param specification: OpenAPI specification. Can be provided either as dict, or as path
            to file.
        :param base_path: Base path to host the API. This overrides the basePath / servers in the
            specification.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A :class:`options.ConnexionOptions` instance with configuration
            options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`
        :param kwargs: Additional keyword arguments to pass to the `add_api` method of the managed
            middlewares. This can be used to pass arguments to middlewares added beyond the default
            ones.

        :return: The Api registered on the middleware application wrapping the framework.
        """
        return self.middleware.add_api(
            specification,
            base_path=base_path,
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            swagger_ui_options=swagger_ui_options,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            **kwargs,
        )

    def add_url_rule(
        self, rule, endpoint: str = None, view_func: t.Callable = None, **options
    ):
        """
        Connects a URL rule.  Works exactly like the `route` decorator.

        Basically this example::

            @app.route('/')
            def index():
                pass

        Is equivalent to the following::

            def index():
                pass
            app.add_url_rule('/', 'index', index)

        Internally`route` invokes `add_url_rule` so if you want to customize the behavior via
        subclassing you only need to change this method.

        :param rule: the URL rule as string.
        :param endpoint: the name of the endpoint for the registered URL rule, which is used for
            reverse lookup. Flask defaults to the name of the view function.
        :param view_func: the function to call when serving a request to the provided endpoint.
        :param options: the options to be forwarded to the underlying `werkzeug.routing.Rule`
            object.  A change to Werkzeug is handling of method options. methods is a list of
            methods this rule should be limited to (`GET`, `POST` etc.).  By default a rule just
            listens for `GET` (and implicitly `HEAD`).
        """

    def route(self, rule: str, **options):
        """
        A decorator that is used to register a view function for a
        given URL rule.  This does the same thing as `add_url_rule`
        but is intended for decorator usage::

            @app.route('/')
            def index():
                return 'Hello World'

        :param rule: the URL rule as string
        :param options: the options to be forwarded to the underlying `werkzeug.routing.Rule`
                        object. A change to Werkzeug is handling of method options. methods is a
                        list of methods this rule should be limited to (`GET`, `POST` etc.).
                        By default a rule just listens for `GET` (and implicitly `HEAD`).
        """

        def decorator(func: t.Callable) -> t.Callable:
            self.add_url_rule(rule, view_func=func, **options)
            return func

        return decorator

    @abc.abstractmethod
    def add_error_handler(
        self, code_or_exception: t.Union[int, t.Type[Exception]], function: t.Callable
    ) -> None:
        """
        Register a callable to handle application errors.

        :param code_or_exception: An exception class or the status code of HTTP exceptions to
            handle.
        :param function: Callable that will handle exception.
        """

    def test_client(self, **kwargs):
        """Creates a test client for this application."""
        return TestClient(self, **kwargs)

    def run(self, import_string: str = None, **kwargs):
        """Run the application using uvicorn.

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param kwargs: kwargs to pass to `uvicorn.run`.
        """
        self.middleware.run(import_string, **kwargs)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        return await self.middleware(scope, receive, send)



import logging

from starlette.exceptions import ExceptionMiddleware as StarletteExceptionMiddleware
from starlette.exceptions import HTTPException
from starlette.requests import Request as StarletteRequest
from starlette.responses import Response
from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.exceptions import InternalServerError, ProblemException, problem

logger = logging.getLogger(__name__)


class ExceptionMiddleware(StarletteExceptionMiddleware):
    """Subclass of starlette ExceptionMiddleware to change handling of HTTP exceptions to
    existing connexion behavior."""

    def __init__(self, next_app: ASGIApp, *args, **kwargs):
        super().__init__(next_app)
        self.add_exception_handler(ProblemException, self.problem_handler)
        self.add_exception_handler(Exception, self.common_error_handler)

    @staticmethod
    def problem_handler(_request: StarletteRequest, exc: ProblemException):
        logger.exception(exc)

        response = exc.to_problem()

        return Response(
            content=response.body,
            status_code=response.status_code,
            media_type=response.mimetype,
            headers=response.headers,
        )

    @staticmethod
    def http_exception(_request: StarletteRequest, exc: HTTPException) -> Response:
        logger.exception(exc)

        headers = exc.headers

        connexion_response = problem(
            title=exc.detail, detail=exc.detail, status=exc.status_code, headers=headers
        )

        return Response(
            content=connexion_response.body,
            status_code=connexion_response.status_code,
            media_type=connexion_response.mimetype,
            headers=connexion_response.headers,
        )

    @staticmethod
    def common_error_handler(_request: StarletteRequest, exc: Exception) -> Response:
        logger.exception(exc)

        response = InternalServerError().to_problem()

        return Response(
            content=response.body,
            status_code=response.status_code,
            media_type=response.mimetype,
            headers=response.headers,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        # Needs to be set so starlette router throws exceptions instead of returning error responses
        scope["app"] = "connexion"
        await super().__call__(scope, receive, send)



"""
This module defines a decorator to convert request parameters to arguments for the view function.
"""
import abc
import asyncio
import builtins
import functools
import inspect
import keyword
import logging
import re
import typing as t
from copy import copy, deepcopy

import inflection

from connexion.context import context, operation
from connexion.frameworks.abstract import Framework
from connexion.http_facts import FORM_CONTENT_TYPES
from connexion.lifecycle import ASGIRequest, WSGIRequest
from connexion.operations import AbstractOperation, Swagger2Operation
from connexion.utils import deep_merge, is_null, is_nullable, make_type

logger = logging.getLogger(__name__)

CONTEXT_NAME = "context_"


class BaseParameterDecorator:
    def __init__(
        self,
        *,
        framework: t.Type[Framework],
        pythonic_params: bool = False,
    ) -> None:
        self.framework = framework
        self.sanitize_fn = pythonic if pythonic_params else sanitized

    def _maybe_get_body(
        self,
        request: t.Union[WSGIRequest, ASGIRequest],
        *,
        arguments: t.List[str],
        has_kwargs: bool,
    ) -> t.Any:
        body_name = self.sanitize_fn(operation.body_name(request.content_type))
        # Pass form contents separately for Swagger2 for backward compatibility with
        # Connexion 2 Checking for body_name is not enough
        if (body_name in arguments or has_kwargs) or (
            request.mimetype in FORM_CONTENT_TYPES
            and isinstance(operation, Swagger2Operation)
        ):
            return request.get_body()
        else:
            return None

    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        raise NotImplementedError


class SyncParameterDecorator(BaseParameterDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        unwrapped_function = unwrap_decorators(function)
        arguments, has_kwargs = inspect_function_arguments(unwrapped_function)

        @functools.wraps(function)
        def wrapper(request: WSGIRequest) -> t.Any:
            request_body = self._maybe_get_body(
                request, arguments=arguments, has_kwargs=has_kwargs
            )

            kwargs = prep_kwargs(
                request,
                request_body=request_body,
                files=request.files(),
                arguments=arguments,
                has_kwargs=has_kwargs,
                sanitize=self.sanitize_fn,
            )

            return function(**kwargs)

        return wrapper


class AsyncParameterDecorator(BaseParameterDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        unwrapped_function = unwrap_decorators(function)
        arguments, has_kwargs = inspect_function_arguments(unwrapped_function)

        @functools.wraps(function)
        async def wrapper(request: ASGIRequest) -> t.Any:
            request_body = self._maybe_get_body(
                request, arguments=arguments, has_kwargs=has_kwargs
            )

            while asyncio.iscoroutine(request_body):
                request_body = await request_body

            kwargs = prep_kwargs(
                request,
                request_body=request_body,
                files=await request.files(),
                arguments=arguments,
                has_kwargs=has_kwargs,
                sanitize=self.sanitize_fn,
            )

            return await function(**kwargs)

        return wrapper


def prep_kwargs(
    request: t.Union[WSGIRequest, ASGIRequest],
    *,
    request_body: t.Any,
    files: t.Dict[str, t.Any],
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
) -> dict:
    kwargs = get_arguments(
        operation,
        path_params=request.path_params,
        query_params=request.query_params,
        body=request_body,
        files=files,
        arguments=arguments,
        has_kwargs=has_kwargs,
        sanitize=sanitize,
        content_type=request.mimetype,
    )

    # optionally convert parameter variable names to un-shadowed, snake_case form
    kwargs = {sanitize(k): v for k, v in kwargs.items()}

    # add context info (e.g. from security decorator)
    for key, value in context.items():
        if has_kwargs or key in arguments:
            kwargs[key] = value
        else:
            logger.debug("Context parameter '%s' not in function arguments", key)
    # attempt to provide the request context to the function
    if CONTEXT_NAME in arguments:
        kwargs[CONTEXT_NAME] = context

    return kwargs


def unwrap_decorators(function: t.Callable) -> t.Callable:
    """Unwrap decorators to return the original function."""
    while hasattr(function, "__wrapped__"):
        function = function.__wrapped__  # type: ignore
    return function


def inspect_function_arguments(function: t.Callable) -> t.Tuple[t.List[str], bool]:
    """
    Returns the list of variables names of a function and if it
    accepts keyword arguments.
    """
    parameters = inspect.signature(function).parameters
    bound_arguments = [
        name
        for name, p in parameters.items()
        if p.kind not in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
    ]
    has_kwargs = any(p.kind == p.VAR_KEYWORD for p in parameters.values())
    return list(bound_arguments), has_kwargs


def snake_and_shadow(name: str) -> str:
    """
    Converts the given name into Pythonic form. Firstly it converts CamelCase names to snake_case. Secondly it looks to
    see if the name matches a known built-in and if it does it appends an underscore to the name.
    :param name: The parameter name
    """
    snake = inflection.underscore(name)
    if snake in builtins.__dict__ or keyword.iskeyword(snake):
        return f"{snake}_"
    return snake


def sanitized(name: str) -> str:
    return name and re.sub(
        "^[^a-zA-Z_]+", "", re.sub("[^0-9a-zA-Z_]", "", re.sub(r"\[(?!])", "_", name))
    )


def pythonic(name: str) -> str:
    name = name and snake_and_shadow(name)
    return sanitized(name)


def get_arguments(
    operation: AbstractOperation,
    *,
    path_params: dict,
    query_params: dict,
    body: t.Any,
    files: dict,
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
    content_type: str,
) -> t.Dict[str, t.Any]:
    """
    get arguments for handler function
    """
    ret = {}
    ret.update(_get_path_arguments(path_params, operation=operation, sanitize=sanitize))
    ret.update(
        _get_query_arguments(
            query_params,
            operation=operation,
            arguments=arguments,
            has_kwargs=has_kwargs,
            sanitize=sanitize,
        )
    )

    if operation.method.upper() in ["PATCH", "POST", "PUT"]:
        ret.update(
            _get_body_argument(
                body,
                operation=operation,
                arguments=arguments,
                has_kwargs=has_kwargs,
                sanitize=sanitize,
                content_type=content_type,
            )
        )
        ret.update(_get_file_arguments(files, arguments, has_kwargs))
    return ret


def _get_path_arguments(
    path_params: dict, *, operation: AbstractOperation, sanitize: t.Callable
) -> dict:
    """
    Extract handler function arguments from path parameters
    """
    kwargs = {}

    path_definitions = {
        parameter["name"]: parameter
        for parameter in operation.parameters
        if parameter["in"] == "path"
    }

    for name, value in path_params.items():
        sanitized_key = sanitize(name)
        if name in path_definitions:
            kwargs[sanitized_key] = _get_val_from_param(value, path_definitions[name])
        else:  # Assume path params mechanism used for injection
            kwargs[sanitized_key] = value
    return kwargs


def _get_val_from_param(value: t.Any, param_definitions: t.Dict[str, dict]) -> t.Any:
    """Cast a value according to its definition in the specification."""
    param_schema = param_definitions.get("schema", param_definitions)

    if is_nullable(param_schema) and is_null(value):
        return None

    if param_schema["type"] == "array":
        type_ = param_schema["items"]["type"]
        format_ = param_schema["items"].get("format")
        return [make_type(part, type_, format_) for part in value]
    else:
        type_ = param_schema["type"]
        format_ = param_schema.get("format")
        return make_type(value, type_, format_)


def _get_query_arguments(
    query_params: dict,
    *,
    operation: AbstractOperation,
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
) -> dict:
    """
    extract handler function arguments from the query parameters
    """
    query_definitions = {
        parameter["name"]: parameter
        for parameter in operation.parameters
        if parameter["in"] == "query"
    }

    default_query_params = _get_query_defaults(query_definitions)

    query_arguments = deepcopy(default_query_params)
    query_arguments = deep_merge(query_arguments, query_params)
    return _query_args_helper(
        query_definitions, query_arguments, arguments, has_kwargs, sanitize
    )


def _get_query_defaults(query_definitions: t.Dict[str, dict]) -> t.Dict[str, t.Any]:
    """Get the default values for the query parameter from the parameter definition."""
    defaults = {}
    for k, v in query_definitions.items():
        try:
            if "default" in v:
                defaults[k] = v["default"]
            elif v["schema"]["type"] == "object":
                defaults[k] = _get_default_obj(v["schema"])
            else:
                defaults[k] = v["schema"]["default"]
        except KeyError:
            pass
    return defaults


def _get_default_obj(schema: dict) -> dict:
    try:
        return deepcopy(schema["default"])
    except KeyError:
        properties = schema.get("properties", {})
        return _build_default_obj_recursive(properties, {})


def _build_default_obj_recursive(properties: dict, default_object: dict) -> dict:
    """takes disparate and nested default keys, and builds up a default object"""
    for name, property_ in properties.items():
        if "default" in property_ and name not in default_object:
            default_object[name] = copy(property_["default"])
        elif property_.get("type") == "object" and "properties" in property_:
            default_object.setdefault(name, {})
            default_object[name] = _build_default_obj_recursive(
                property_["properties"], default_object[name]
            )
    return default_object


def _query_args_helper(
    query_definitions: dict,
    query_arguments: dict,
    function_arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
) -> dict:
    result = {}
    for key, value in query_arguments.items():
        sanitized_key = sanitize(key)
        if not has_kwargs and sanitized_key not in function_arguments:
            logger.debug(
                "Query Parameter '%s' (sanitized: '%s') not in function arguments",
                key,
                sanitized_key,
            )
        else:
            logger.debug(
                "Query Parameter '%s' (sanitized: '%s') in function arguments",
                key,
                sanitized_key,
            )
            try:
                query_defn = query_definitions[key]
            except KeyError:  # pragma: no cover
                logger.error(
                    "Function argument '%s' (non-sanitized: %s) not defined in specification",
                    sanitized_key,
                    key,
                )
            else:
                logger.debug("%s is a %s", key, query_defn)
                result.update({sanitized_key: _get_val_from_param(value, query_defn)})
    return result


def _get_body_argument(
    body: t.Any,
    *,
    operation: AbstractOperation,
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
    content_type: str,
) -> dict:
    if len(arguments) <= 0 and not has_kwargs:
        return {}

    body_name = sanitize(operation.body_name(content_type))

    if content_type in FORM_CONTENT_TYPES:
        result = _get_body_argument_form(
            body, operation=operation, content_type=content_type
        )

        # Unpack form values for Swagger for compatibility with Connexion 2 behavior
        if content_type in FORM_CONTENT_TYPES and isinstance(
            operation, Swagger2Operation
        ):
            if has_kwargs:
                return result
            else:
                return {
                    sanitize(name): value
                    for name, value in result.items()
                    if sanitize(name) in arguments
                }
    else:
        result = _get_body_argument_json(
            body, operation=operation, content_type=content_type
        )

    if body_name in arguments or has_kwargs:
        return {body_name: result}

    return {}


def _get_body_argument_json(
    body: t.Any, *, operation: AbstractOperation, content_type: str
) -> t.Any:
    # if the body came in null, and the schema says it can be null, we decide
    # to include no value for the body argument, rather than the default body
    if is_nullable(operation.body_schema(content_type)) and is_null(body):
        return None

    if body is None:
        default_body = operation.body_schema(content_type).get("default", {})
        return deepcopy(default_body)

    return body


def _get_body_argument_form(
    body: dict, *, operation: AbstractOperation, content_type: str
) -> dict:
    # now determine the actual value for the body (whether it came in or is default)
    default_body = operation.body_schema(content_type).get("default", {})
    body_props = {
        k: {"schema": v}
        for k, v in operation.body_schema(content_type).get("properties", {}).items()
    }

    # by OpenAPI specification `additionalProperties` defaults to `true`
    # see: https://github.com/OAI/OpenAPI-Specification/blame/3.0.2/versions/3.0.2.md#L2305
    additional_props = operation.body_schema().get("additionalProperties", True)

    body_arg = deepcopy(default_body)
    body_arg.update(body or {})

    if body_props or additional_props:
        return _get_typed_body_values(body_arg, body_props, additional_props)

    return {}


def _get_typed_body_values(body_arg, body_props, additional_props):
    """
    Return a copy of the provided body_arg dictionary
    whose values will have the appropriate types
    as defined in the provided schemas.

    :type body_arg: type dict
    :type body_props: dict
    :type additional_props: dict|bool
    :rtype: dict
    """
    additional_props_defn = (
        {"schema": additional_props} if isinstance(additional_props, dict) else None
    )
    res = {}

    for key, value in body_arg.items():
        try:
            prop_defn = body_props[key]
            res[key] = _get_val_from_param(value, prop_defn)
        except KeyError:  # pragma: no cover
            if not additional_props:
                logger.error(f"Body property '{key}' not defined in body schema")
                continue
            if additional_props_defn is not None:
                value = _get_val_from_param(value, additional_props_defn)
            res[key] = value

    return res


def _get_file_arguments(files, arguments, has_kwargs=False):
    return {k: v for k, v in files.items() if k in arguments or has_kwargs}



import logging
import typing as t
from collections import defaultdict

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.exceptions import ProblemException
from connexion.lifecycle import ASGIRequest
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation
from connexion.security import SecurityHandlerFactory

logger = logging.getLogger("connexion.middleware.security")


class SecurityOperation:
    def __init__(
        self,
        next_app: ASGIApp,
        *,
        security_handler_factory: SecurityHandlerFactory,
        security: list,
        security_schemes: dict,
    ):
        self.next_app = next_app
        self.security_handler_factory = security_handler_factory
        self.security = security
        self.security_schemes = security_schemes
        self.verification_fn = self._get_verification_fn()

    @classmethod
    def from_operation(
        cls,
        operation: AbstractOperation,
        *,
        next_app: ASGIApp,
        security_handler_factory: SecurityHandlerFactory,
    ) -> "SecurityOperation":
        return cls(
            next_app=next_app,
            security_handler_factory=security_handler_factory,
            security=operation.security,
            security_schemes=operation.security_schemes,
        )

    def _get_verification_fn(self):
        logger.debug("... Security: %s", self.security, extra=vars(self))
        if not self.security:
            return self.security_handler_factory.security_passthrough

        auth_funcs = []
        for security_req in self.security:
            if not security_req:
                auth_funcs.append(self.security_handler_factory.verify_none())
                continue

            sec_req_funcs = {}
            oauth = False
            for scheme_name, required_scopes in security_req.items():
                security_scheme = self.security_schemes[scheme_name]

                if security_scheme["type"] == "oauth2":
                    if oauth:
                        logger.warning(
                            "... multiple OAuth2 security schemes in AND fashion not supported",
                            extra=vars(self),
                        )
                        break
                    oauth = True
                    token_info_func = self.security_handler_factory.get_tokeninfo_func(
                        security_scheme
                    )
                    scope_validate_func = (
                        self.security_handler_factory.get_scope_validate_func(
                            security_scheme
                        )
                    )
                    if not token_info_func:
                        logger.warning("... x-tokenInfoFunc missing", extra=vars(self))
                        break

                    sec_req_funcs[
                        scheme_name
                    ] = self.security_handler_factory.verify_oauth(
                        token_info_func, scope_validate_func, required_scopes
                    )

                # Swagger 2.0
                elif security_scheme["type"] == "basic":
                    basic_info_func = self.security_handler_factory.get_basicinfo_func(
                        security_scheme
                    )
                    if not basic_info_func:
                        logger.warning("... x-basicInfoFunc missing", extra=vars(self))
                        break

                    sec_req_funcs[
                        scheme_name
                    ] = self.security_handler_factory.verify_basic(basic_info_func)

                # OpenAPI 3.0.0
                elif security_scheme["type"] == "http":
                    scheme = security_scheme["scheme"].lower()
                    if scheme == "basic":
                        basic_info_func = (
                            self.security_handler_factory.get_basicinfo_func(
                                security_scheme
                            )
                        )
                        if not basic_info_func:
                            logger.warning(
                                "... x-basicInfoFunc missing", extra=vars(self)
                            )
                            break

                        sec_req_funcs[
                            scheme_name
                        ] = self.security_handler_factory.verify_basic(basic_info_func)
                    elif scheme == "bearer":
                        bearer_info_func = (
                            self.security_handler_factory.get_bearerinfo_func(
                                security_scheme
                            )
                        )
                        if not bearer_info_func:
                            logger.warning(
                                "... x-bearerInfoFunc missing", extra=vars(self)
                            )
                            break
                        sec_req_funcs[
                            scheme_name
                        ] = self.security_handler_factory.verify_bearer(
                            bearer_info_func
                        )
                    else:
                        logger.warning(
                            "... Unsupported http authorization scheme %s" % scheme,
                            extra=vars(self),
                        )
                        break

                elif security_scheme["type"] == "apiKey":
                    scheme = security_scheme.get("x-authentication-scheme", "").lower()
                    if scheme == "bearer":
                        bearer_info_func = (
                            self.security_handler_factory.get_bearerinfo_func(
                                security_scheme
                            )
                        )
                        if not bearer_info_func:
                            logger.warning(
                                "... x-bearerInfoFunc missing", extra=vars(self)
                            )
                            break
                        sec_req_funcs[
                            scheme_name
                        ] = self.security_handler_factory.verify_bearer(
                            bearer_info_func
                        )
                    else:
                        apikey_info_func = (
                            self.security_handler_factory.get_apikeyinfo_func(
                                security_scheme
                            )
                        )
                        if not apikey_info_func:
                            logger.warning(
                                "... x-apikeyInfoFunc missing", extra=vars(self)
                            )
                            break

                        sec_req_funcs[
                            scheme_name
                        ] = self.security_handler_factory.verify_api_key(
                            apikey_info_func,
                            security_scheme["in"],
                            security_scheme["name"],
                        )

                else:
                    logger.warning(
                        "... Unsupported security scheme type %s"
                        % security_scheme["type"],
                        extra=vars(self),
                    )
                    break
            else:
                # No break encountered: no missing funcs
                if len(sec_req_funcs) == 1:
                    (func,) = sec_req_funcs.values()
                    auth_funcs.append(func)
                else:
                    auth_funcs.append(
                        self.security_handler_factory.verify_multiple_schemes(
                            sec_req_funcs
                        )
                    )

        return self.security_handler_factory.verify_security(auth_funcs)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = ASGIRequest(scope)
        await self.verification_fn(request)
        await self.next_app(scope, receive, send)


class SecurityAPI(RoutedAPI[SecurityOperation]):
    def __init__(self, *args, auth_all_paths: bool = False, **kwargs):
        super().__init__(*args, **kwargs)

        self.security_handler_factory = SecurityHandlerFactory()

        if auth_all_paths:
            self.add_auth_on_not_found()
        else:
            self.operations: t.MutableMapping[str, SecurityOperation] = {}

        self.add_paths()

    def add_auth_on_not_found(self) -> None:
        """Register a default SecurityOperation for routes that are not found."""
        default_operation = self.make_operation(self.specification)
        self.operations = defaultdict(lambda: default_operation)

    def make_operation(self, operation: AbstractOperation) -> SecurityOperation:
        return SecurityOperation.from_operation(
            operation,
            next_app=self.next_app,
            security_handler_factory=self.security_handler_factory,
        )


class SecurityMiddleware(RoutedMiddleware[SecurityAPI]):
    """Middleware to check if operation is accessible on scope."""

    api_cls = SecurityAPI


class MissingSecurityOperation(ProblemException):
    pass



import abc
import collections.abc
import functools
import logging
import types
import typing as t
from enum import Enum

from connexion.context import operation
from connexion.datastructures import NoContent
from connexion.exceptions import NonConformingResponseHeaders
from connexion.frameworks.abstract import Framework
from connexion.lifecycle import ConnexionResponse
from connexion.utils import is_json_mimetype

logger = logging.getLogger(__name__)


class BaseResponseDecorator:
    def __init__(self, *, framework: t.Type[Framework], jsonifier):
        self.framework = framework
        self.jsonifier = jsonifier

    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        raise NotImplementedError

    def build_framework_response(self, handler_response):
        data, status_code, headers = self._unpack_handler_response(handler_response)
        content_type = self._deduct_content_type(data, headers)
        if not self.framework.is_framework_response(data):
            data, status_code = self._prepare_body_and_status_code(
                data, status_code=status_code, mimetype=content_type
            )
        return self.framework.build_response(
            data, content_type=content_type, status_code=status_code, headers=headers
        )

    @staticmethod
    def _deduct_content_type(data: t.Any, headers: dict) -> str:
        """Deduct the response content type from the returned data, headers and operation spec.

        :param data: Response data
        :param headers: Headers returned by the handler.

        :return: Deducted content type

        :raises: NonConformingResponseHeaders if content type cannot be deducted.
        """
        content_type = headers.get("Content-Type")

        # TODO: don't default
        produces = list(set(operation.produces))
        if data is not None and not produces:
            produces = ["application/json"]

        if content_type:
            if content_type not in produces:
                raise NonConformingResponseHeaders(
                    f"Returned content type ({content_type}) is not defined in operation spec "
                    f"({operation.produces})."
                )
        else:
            if not produces:
                # Produces can be empty/ for empty responses
                pass
            elif len(produces) == 1:
                content_type = produces[0]
            elif isinstance(data, str) and "text/plain" in produces:
                content_type = "text/plain"
            elif (
                isinstance(data, bytes)
                or isinstance(data, (types.GeneratorType, collections.abc.Iterator))
            ) and "application/octet-stream" in produces:
                content_type = "application/octet-stream"
            else:
                raise NonConformingResponseHeaders(
                    "Multiple response content types are defined in the operation spec, but the "
                    "handler response did not specify which one to return."
                )

        return content_type

    def _prepare_body_and_status_code(
        self, data, *, status_code: int = None, mimetype: str
    ) -> tuple:
        if data is NoContent:
            data = None

        if status_code is None:
            if data is None:
                status_code = 204
            else:
                status_code = 200

        if data is not None:
            body = self._serialize_data(data, mimetype)
        else:
            body = data

        return body, status_code

    def _serialize_data(self, data: t.Any, mimetype: str) -> t.Any:
        if is_json_mimetype(mimetype):
            return self.jsonifier.dumps(data)
        return data

    @staticmethod
    def _unpack_handler_response(
        handler_response: t.Union[str, bytes, dict, list, tuple]
    ) -> t.Tuple[t.Union[str, bytes, dict, list, None], t.Optional[int], dict]:
        """Unpack the handler response into data, status_code and headers.

        :param handler_response: The response returned from the handler function if it was not a
            response class.

        :return: A tuple of data, status_code and headers
        """
        data, status_code, headers = None, None, {}

        if not isinstance(handler_response, tuple):
            data = handler_response

        elif len(handler_response) == 1:
            (data,) = handler_response

        elif len(handler_response) == 2:
            data, status_code_or_headers = handler_response
            if isinstance(status_code_or_headers, int):
                status_code = status_code_or_headers
            elif isinstance(status_code_or_headers, Enum) and isinstance(
                status_code_or_headers.value, int
            ):
                status_code = status_code_or_headers.value
            else:
                headers = status_code_or_headers

        elif len(handler_response) == 3:
            data, status_code, headers = handler_response

        else:
            raise TypeError(
                "The view function did not return a valid response tuple."
                " The tuple must have the form (body), (body, status, headers),"
                " (body, status), or (body, headers)."
            )

        return data, status_code, headers


class SyncResponseDecorator(BaseResponseDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            """
            This method converts a handler response to a framework response.
            The handler response can be a ConnexionResponse, a framework response, a tuple or an
            object.
            """
            handler_response = function(*args, **kwargs)
            if self.framework.is_framework_response(handler_response):
                return handler_response
            elif isinstance(handler_response, ConnexionResponse):
                return self.framework.connexion_to_framework_response(handler_response)
            else:
                return self.build_framework_response(handler_response)

        return wrapper


class AsyncResponseDecorator(BaseResponseDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            """
            This method converts a handler response to a framework response.
            The handler response can be a ConnexionResponse, a framework response, a tuple or an
            object.
            """
            handler_response = await function(*args, **kwargs)
            if self.framework.is_framework_response(handler_response):
                return handler_response
            elif isinstance(handler_response, ConnexionResponse):
                return self.framework.connexion_to_framework_response(handler_response)
            else:
                return self.build_framework_response(handler_response)

        return wrapper
